@@ -29,6 +29,8 @@
 #include <sys/socket.h>
 #include <arpa/inet.h>
 #include <time.h>
+#include <errno.h>
+#include <limits.h>
 
 #include "private.h"
 #include "time.h"
@@ -36,10 +38,12 @@
 #include "dhcp.h"
 #include "dhcp-private.h"
 #include "queue.h"
-#include "util.h"
+#include "useful.h"
 #include "strv.h"
 #include "timeout.h"
 #include "acd.h"
+#include "log.h"
+#include "util.h"
 
 /* 8 hours */
 #define DEFAULT_DHCP_LEASE_SEC (8*60*60)
@@ -83,30 +87,41 @@ struct l_dhcp_server {
 	struct dhcp_transport *transport;
 
 	struct l_acd *acd;
+
+	bool authoritative : 1;
+	bool rapid_commit : 1;
 };
 
 #define MAC "%02x:%02x:%02x:%02x:%02x:%02x"
 #define MAC_STR(a) a[0], a[1], a[2], a[3], a[4], a[5]
 
-#define IP_STR(uint_ip) \
-({ \
-	struct in_addr _in; \
-	char *_out; \
-	_in.s_addr = uint_ip; \
-	_out = inet_ntoa(_in); \
-	_out; \
-})
+#define NIPQUAD_FMT "%u.%u.%u.%u"
+#define NIPQUAD(u32_ip)	((unsigned char *) &u32_ip)[0], \
+			((unsigned char *) &u32_ip)[1], \
+			((unsigned char *) &u32_ip)[2], \
+			((unsigned char *) &u32_ip)[3]
 
 #define SERVER_DEBUG(fmt, args...)					\
 	l_util_debug(server->debug_handler, server->debug_data,		\
 			"%s:%i " fmt, __func__, __LINE__, ## args)
 
-static bool is_expired_lease(struct l_dhcp_lease *lease)
+static uint64_t get_lease_expiry_time(const struct l_dhcp_lease *lease)
 {
-	if (lease->lifetime < l_time_to_secs(l_time_now()))
-		return true;
+	return lease->bound_time + lease->lifetime * L_USEC_PER_SEC;
+}
 
-	return false;
+static bool is_expired_lease(const struct l_dhcp_lease *lease)
+{
+	return !l_time_after(get_lease_expiry_time(lease), l_time_now());
+}
+
+static bool match_lease_client_id(const void *data, const void *user_data)
+{
+	const struct l_dhcp_lease *lease = data;
+	const uint8_t *client_id = user_data;
+
+	return lease->client_id &&
+		!memcmp(lease->client_id, client_id, client_id[0] + 1);
 }
 
 static bool match_lease_mac(const void *data, const void *user_data)
@@ -117,24 +132,55 @@ static bool match_lease_mac(const void *data, const void *user_data)
 	return !memcmp(lease->mac, mac, 6);
 }
 
-static struct l_dhcp_lease *find_lease_by_mac(struct l_dhcp_server *server,
-						const uint8_t *mac)
+static bool match_lease_ip(const void *data, const void *user_data)
 {
-	return l_queue_find(server->lease_list, match_lease_mac, mac);
+	const struct l_dhcp_lease *lease = data;
+
+	return lease->address == L_PTR_TO_UINT(user_data);
 }
 
-static void remove_lease(struct l_dhcp_server *server,
-				struct l_dhcp_lease *lease)
+static struct l_dhcp_lease *find_lease_by_ip(struct l_queue *lease_list,
+						uint32_t nip)
 {
-	l_queue_remove(server->lease_list, lease);
+	return l_queue_find(lease_list, match_lease_ip, L_UINT_TO_PTR(nip));
+}
 
-	_dhcp_lease_free(lease);
+static struct l_dhcp_lease *find_lease_by_id(struct l_queue *lease_list,
+						const uint8_t *client_id,
+						const uint8_t *mac)
+{
+	if (client_id)
+		return l_queue_find(lease_list, match_lease_client_id,
+					client_id);
+
+	return l_queue_find(lease_list, match_lease_mac, mac);
+}
+
+static struct l_dhcp_lease *find_lease_by_id_and_ip(struct l_queue *lease_list,
+						const uint8_t *client_id,
+						const uint8_t *mac,
+						uint32_t ip)
+{
+	struct l_dhcp_lease *lease = find_lease_by_ip(lease_list, ip);
+
+	if (!lease)
+		return NULL;
+
+	if (client_id) {
+		if (!match_lease_client_id(lease, client_id))
+			return NULL;
+	} else {
+		if (!match_lease_mac(lease, mac))
+			return NULL;
+	}
+
+	return lease;
 }
 
 /* Clear the old lease and create the new one */
 static int get_lease(struct l_dhcp_server *server, uint32_t yiaddr,
-				const uint8_t *mac,
-				struct l_dhcp_lease **lease_out)
+			const uint8_t *client_id, const uint8_t *mac,
+			struct l_dhcp_lease **lease_out)
 {
 	struct l_dhcp_lease *lease;
 
@@ -153,13 +199,17 @@ static int get_lease(struct l_dhcp_server *server, uint32_t yiaddr,
 	if (l_memeqzero(mac, ETH_ALEN))
 		return -ENXIO;
 
-	lease = find_lease_by_mac(server, mac);
-
+	lease = find_lease_by_ip(server->lease_list, yiaddr);
 	if (lease) {
 		l_queue_remove(server->lease_list, lease);
-
 		*lease_out = lease;
+		return 0;
+	}
 
+	lease = find_lease_by_ip(server->expired_list, yiaddr);
+	if (lease && lease->address == yiaddr) {
+		l_queue_remove(server->expired_list, lease);
+		*lease_out = lease;
 		return 0;
 	}
 
@@ -168,11 +218,12 @@ static int get_lease(struct l_dhcp_server *server, uint32_t yiaddr,
 	return 0;
 }
 
-static int compare_lifetime_or_offering(const void *a, const void *b,
+static int compare_expiry_or_offering(const void *a, const void *b,
 					void *user_data)
 {
 	const struct l_dhcp_lease *lease1 = a;
 	const struct l_dhcp_lease *lease2 = b;
+	int64_t diff;
 
 	/*
 	 * Ensures offered but not active leases stay at the head of the queue.
@@ -181,7 +232,9 @@ static int compare_lifetime_or_offering(const void *a, const void *b,
 	if (lease1->offering)
 		return 1;
 
-	return lease2->lifetime - lease1->lifetime;
+	diff = (int64_t) lease2->bound_time - lease1->bound_time +
+		((int64_t) lease2->lifetime - lease1->lifetime) * L_USEC_PER_SEC;
+	return diff >= 0 ? diff > 0 ? 1 : 0 : -1;
 }
 
 static void lease_expired_cb(struct l_timeout *timeout, void *user_data);
@@ -190,7 +243,9 @@ static void set_next_expire_timer(struct l_dhcp_server *server,
 					struct l_dhcp_lease *expired)
 {
 	struct l_dhcp_lease *next;
-	unsigned int next_timeout;
+	uint64_t expiry;
+	uint64_t now;
+	uint64_t next_timeout;
 
 	/*
 	 * If this is an expiring lease put it into the expired queue, removing
@@ -199,26 +254,33 @@ static void set_next_expire_timer(struct l_dhcp_server *server,
 	if (expired) {
 		l_queue_remove(server->lease_list, expired);
 
-		if (l_queue_length(server->expired_list) > server->max_expired)
-			_dhcp_lease_free(l_queue_pop_head(
+		if (!expired->offering) {
+			if (l_queue_length(server->expired_list) >
+					server->max_expired)
+				_dhcp_lease_free(l_queue_pop_head(
 							server->expired_list));
 
-		l_queue_push_tail(server->expired_list, expired);
+			l_queue_push_tail(server->expired_list, expired);
+		} else
+			_dhcp_lease_free(expired);
 	}
 
 	next = l_queue_peek_tail(server->lease_list);
-	if (!next || next->offering) {
+	if (!next) {
+		l_timeout_remove(server->next_expire);
 		server->next_expire = NULL;
 		return;
 	}
 
-	next_timeout = l_time_to_secs(l_time_diff(l_time_now(),
-							next->lifetime));
+	expiry = get_lease_expiry_time(next);
+	now = l_time_now();
+	next_timeout = l_time_after(expiry, now) ?
+		l_time_to_msecs(expiry - now) : 0;
 
 	if (server->next_expire)
-		l_timeout_modify(server->next_expire, next_timeout);
+		l_timeout_modify_ms(server->next_expire, next_timeout ?: 1);
 	else
-		server->next_expire = l_timeout_create(server->lease_seconds,
+		server->next_expire = l_timeout_create(next_timeout ?: 1,
 							lease_expired_cb,
 							server, NULL);
 }
@@ -226,9 +288,9 @@ static void set_next_expire_timer(struct l_dhcp_server *server,
 static void lease_expired_cb(struct l_timeout *timeout, void *user_data)
 {
 	struct l_dhcp_server *server = user_data;
-	struct l_dhcp_lease *lease = l_queue_peek_tail(server->expired_list);
+	struct l_dhcp_lease *lease = l_queue_peek_tail(server->lease_list);
 
-	if (server->event_handler)
+	if (!lease->offering && server->event_handler)
 		server->event_handler(server, L_DHCP_SERVER_EVENT_LEASE_EXPIRED,
 					server->user_data, lease);
 
@@ -236,62 +298,79 @@ static void lease_expired_cb(struct l_timeout *timeout, void *user_data)
 }
 
 static struct l_dhcp_lease *add_lease(struct l_dhcp_server *server,
-					bool offering, const uint8_t *chaddr,
-					uint32_t yiaddr)
+					bool offering, const uint8_t *client_id,
+					const uint8_t *chaddr, uint32_t yiaddr)
 {
 	struct l_dhcp_lease *lease = NULL;
 	int ret;
 
-	ret = get_lease(server, yiaddr, chaddr, &lease);
+	ret = get_lease(server, yiaddr, client_id, chaddr, &lease);
 	if (ret != 0)
 		return NULL;
 
+	l_free(lease->client_id);
 	memset(lease, 0, sizeof(*lease));
 
 	memcpy(lease->mac, chaddr, ETH_ALEN);
 	lease->address = yiaddr;
+	lease->subnet_mask = server->netmask;
+	lease->router = server->gateway;
+
+	if (server->dns_list) {
+		unsigned int i;
+
+		for (i = 0; server->dns_list[i]; i++);
+		lease->dns = l_memdup(server->dns_list, (i + 1) * 4);
+	}
+
+	if (client_id)
+		lease->client_id = l_memdup(client_id, client_id[0] + 1);
 
 	lease->offering = offering;
-	lease->lifetime = l_time_to_secs(l_time_now());
+	lease->bound_time = l_time_now();
 
 	if (!offering) {
-		lease->lifetime += server->lease_seconds;
+		lease->lifetime = server->lease_seconds;
 
 		/*
 		 * Insert into queue by lifetime (skipping any offered leases
 		 * at the head)
 		 */
 		l_queue_insert(server->lease_list, lease,
-					compare_lifetime_or_offering, NULL);
-		/*
-		 * This is a new (or renewed lease) so pass NULL for expired so
-		 * the queue's are not modified, only the next_expired timer
-		 */
-		set_next_expire_timer(server, NULL);
+					compare_expiry_or_offering, NULL);
 	} else {
-		lease->lifetime += OFFER_TIME;
+		lease->lifetime = OFFER_TIME;
 		/* Push offered leases to head, active leases after those */
 		l_queue_push_head(server->lease_list, lease);
 	}
 
-	SERVER_DEBUG("added lease IP %s for "MAC " lifetime=%u",
-			IP_STR(yiaddr), MAC_STR(chaddr),
+	/*
+	 * This is a new (or renewed) lease so pass NULL for expired so the
+	 * queues are not modified, only the next_expire timer.
+	 */
+	set_next_expire_timer(server, NULL);
+
+	SERVER_DEBUG("added lease IP "NIPQUAD_FMT " for "MAC " lifetime=%u",
+			NIPQUAD(yiaddr), MAC_STR(chaddr),
 			server->lease_seconds);
 
 	return lease;
 }
 
+static bool remove_lease(struct l_dhcp_server *server,
+				struct l_dhcp_lease *lease)
+{
+	if (!l_queue_remove(server->lease_list, lease))
+		return false;
+
+	_dhcp_lease_free(lease);
+	set_next_expire_timer(server, NULL);
+	return true;
+}
+
 static void lease_release(struct l_dhcp_server *server,
 			struct l_dhcp_lease *lease)
 {
-	/*
-	 * If the client released the lease after the server timeout expired
-	 * there is nothing to do. Otherwise the client is releasing the
-	 * lease early which may require re-setting the lease expire timer
-	 */
-	if (is_expired_lease(lease))
-		return;
-
 	if (server->event_handler)
 		server->event_handler(server, L_DHCP_SERVER_EVENT_LEASE_EXPIRED,
 					server->user_data, lease);
@@ -299,19 +378,6 @@ static void lease_release(struct l_dhcp_server *server,
 	set_next_expire_timer(server, lease);
 }
 
-static bool match_lease_ip(const void *data, const void *user_data)
-{
-	const struct l_dhcp_lease *lease = data;
-
-	return lease->address == L_PTR_TO_UINT(user_data);
-}
-
-static struct l_dhcp_lease *find_lease_by_ip(struct l_queue *lease_list,
-						uint32_t nip)
-{
-	return l_queue_find(lease_list, match_lease_ip, L_UINT_TO_PTR(nip));
-}
-
 static bool check_requested_ip(struct l_dhcp_server *server,
 				uint32_t requested_nip)
 {
@@ -323,7 +389,10 @@ static bool check_requested_ip(struct l_dhcp_server *server,
 	if (ntohl(requested_nip) < server->start_ip)
 		return false;
 
-	if (htonl(requested_nip) > server->end_ip)
+	if (ntohl(requested_nip) > server->end_ip)
+		return false;
+
+	if (requested_nip == server->address)
 		return false;
 
 	lease = find_lease_by_ip(server->lease_list, requested_nip);
@@ -361,6 +430,9 @@ static uint32_t find_free_or_expired_ip(struct l_dhcp_server *server,
 		if ((ip_addr & 0xff) == 0xff)
 			continue;
 
+		if (ip_nl == server->address)
+			continue;
+
 		/*
 		 * Search both active and expired leases. If this exausts all
 		 * IP's in the range pop the expired list (oldest expired lease)
@@ -372,7 +444,7 @@ static uint32_t find_free_or_expired_ip(struct l_dhcp_server *server,
 			continue;
 
 		lease = find_lease_by_ip(server->expired_list, ip_nl);
-		if (lease)
+		if (lease && memcmp(lease->mac, safe_mac, ETH_ALEN))
 			continue;
 
 		if (arp_check(ip_nl, safe_mac))
@@ -383,7 +455,9 @@ static uint32_t find_free_or_expired_ip(struct l_dhcp_server *server,
 	if (!lease)
 		return 0;
 
-	return lease->address;
+	ip_addr = lease->address;
+	_dhcp_lease_free(lease);
+	return ip_addr;
 }
 
 static void server_message_init(struct l_dhcp_server *server,
@@ -397,6 +471,71 @@ static void server_message_init(struct l_dhcp_server *server,
 	reply->ciaddr = client_msg->ciaddr;
 }
 
+static bool server_message_send(struct l_dhcp_server *server,
+				struct dhcp_message *reply, size_t len,
+				uint8_t type)
+{
+	uint32_t daddr;
+	uint16_t dport;
+	const uint8_t *dest_mac;
+
+	/*
+	 * RFC2131 Section 4.1: "If the 'giaddr' field in a DHCP message from
+	 * a client is non-zero, the server sends any return messages to the
+	 * 'DHCP server' port on the BOOTP relay agent whose address appears
+	 * in 'giaddr'. If the 'giaddr' field is zero and the 'ciaddr' field
+	 * is nonzero, then the server unicasts DHCPOFFER and DHCPACK messages
+	 * to the address in 'ciaddr'.  If 'giaddr' is zero and 'ciaddr' is
+	 * zero, and the broadcast bit is set, then the server broadcasts
+	 * DHCPOFFER and DHCPACK messages to 0xffffffff. If the broadcast bit
+	 * is not set and 'giaddr' is zero and 'ciaddr' is zero, then the
+	 * server unicasts DHCPOFFER and DHCPACK messages to the client's
+	 * hardware address and 'yiaddr' address.  In all cases, when 'giaddr'
+	 * is zero, the server broadcasts any DHCPNAK messages to 0xffffffff."
+	 *
+	 * 4.3.2: "If 'giaddr' is set in the DHCPREQUEST message, the client
+	 * is on a different subnet.  The server MUST set the broadcast bit in
+	 * the DHCPNAK, so that the relay agent will broadcast the DHCPNAK to
+	 * the client, because the client may not have a correct network
+	 * address or subnet mask, and the client may not be answering ARP
+	 * requests."
+	 */
+	if (reply->giaddr) {
+		dport = DHCP_PORT_SERVER;
+		daddr = reply->giaddr;
+		dest_mac = reply->chaddr;
+
+		if (type == DHCP_MESSAGE_TYPE_NAK)
+			reply->flags |= L_CPU_TO_BE16(DHCP_FLAG_BROADCAST);
+	} else {
+		dport = DHCP_PORT_CLIENT;
+
+		if (type == DHCP_MESSAGE_TYPE_NAK) {
+			daddr = 0xffffffff;
+			dest_mac = MAC_BCAST_ADDR;
+		} else if (reply->ciaddr) {
+			daddr = reply->ciaddr;
+			dest_mac = reply->chaddr;
+		} else if (L_BE16_TO_CPU(reply->flags) & DHCP_FLAG_BROADCAST) {
+			daddr = 0xffffffff;
+			dest_mac = MAC_BCAST_ADDR;
+		} else {
+			daddr = reply->yiaddr;
+			dest_mac = reply->chaddr;
+		}
+	}
+
+	if (server->transport->l2_send(server->transport, server->address,
+					DHCP_PORT_SERVER, daddr, dport,
+					dest_mac, reply, len) < 0) {
+		SERVER_DEBUG("Failed to send %s",
+				_dhcp_message_type_to_string(type));
+		return false;
+	}
+
+	return true;
+}
+
 static void add_server_options(struct l_dhcp_server *server,
 				struct dhcp_message_builder *builder)
 {
@@ -419,9 +558,20 @@ static void add_server_options(struct l_dhcp_server *server,
 	}
 }
 
+/* Copy the client identifier option from the client message per RFC6842 */
+static void copy_client_id(struct dhcp_message_builder *builder,
+				const uint8_t *client_id)
+{
+	if (client_id)
+		_dhcp_message_builder_append(builder,
+						DHCP_OPTION_CLIENT_IDENTIFIER,
+						client_id[0], client_id + 1);
+}
+
 static void send_offer(struct l_dhcp_server *server,
 			const struct dhcp_message *client_msg,
-			struct l_dhcp_lease *lease, uint32_t requested_ip)
+			struct l_dhcp_lease *lease, uint32_t requested_ip,
+			const uint8_t *client_id)
 {
 	struct dhcp_message_builder builder;
 	size_t len = sizeof(struct dhcp_message) + DHCP_MIN_OPTIONS_SIZE;
@@ -439,13 +589,14 @@ static void send_offer(struct l_dhcp_server *server,
 							client_msg->chaddr);
 
 	if (!reply->yiaddr) {
-		SERVER_DEBUG("Could not find lease or send offer");
+		SERVER_DEBUG("No free IP addresses, OFFER abandoned");
 		return;
 	}
 
-	lease = add_lease(server, true, client_msg->chaddr, reply->yiaddr);
+	lease = add_lease(server, true, client_id, client_msg->chaddr,
+				reply->yiaddr);
 	if (!lease) {
-		SERVER_DEBUG("No free IP addresses, OFFER abandoned");
+		SERVER_DEBUG("add_lease() failed");
 		return;
 	}
 
@@ -462,21 +613,19 @@ static void send_offer(struct l_dhcp_server *server,
 					4, &server->address);
 
 	add_server_options(server, &builder);
+	copy_client_id(&builder, client_id);
 
 	_dhcp_message_builder_finalize(&builder, &len);
 
-	SERVER_DEBUG("Sending OFFER of %s to "MAC, IP_STR(reply->yiaddr),
-			MAC_STR(reply->chaddr));
+	SERVER_DEBUG("Sending OFFER of "NIPQUAD_FMT " to "MAC,
+			NIPQUAD(reply->yiaddr),	MAC_STR(reply->chaddr));
 
-	if (server->transport->l2_send(server->transport, server->address,
-					DHCP_PORT_SERVER,
-					reply->ciaddr, DHCP_PORT_CLIENT,
-					reply->chaddr, reply, len) < 0)
-		SERVER_DEBUG("Failed to send OFFER");
+	server_message_send(server, reply, len, DHCP_MESSAGE_TYPE_OFFER);
 }
 
 static void send_inform(struct l_dhcp_server *server,
-				const struct dhcp_message *client_msg)
+				const struct dhcp_message *client_msg,
+				const uint8_t *client_id)
 {
 	struct dhcp_message_builder builder;
 	size_t len = sizeof(struct dhcp_message) + DHCP_MIN_OPTIONS_SIZE;
@@ -489,18 +638,16 @@ static void send_inform(struct l_dhcp_server *server,
 	_dhcp_message_builder_init(&builder, reply, len, DHCP_MESSAGE_TYPE_ACK);
 
 	add_server_options(server, &builder);
+	copy_client_id(&builder, client_id);
 
 	_dhcp_message_builder_finalize(&builder, &len);
 
-	if (server->transport->l2_send(server->transport, server->address,
-					DHCP_PORT_SERVER, reply->ciaddr,
-					DHCP_PORT_CLIENT, reply->chaddr,
-					reply, len) < 0)
-		SERVER_DEBUG("Failed to send INFORM");
+	server_message_send(server, reply, len, DHCP_MESSAGE_TYPE_INFORM);
 }
 
 static void send_nak(struct l_dhcp_server *server,
-			const struct dhcp_message *client_msg)
+			const struct dhcp_message *client_msg,
+			const uint8_t *client_id)
 {
 	struct dhcp_message_builder builder;
 	size_t len = sizeof(struct dhcp_message) + DHCP_MIN_OPTIONS_SIZE;
@@ -511,24 +658,22 @@ static void send_nak(struct l_dhcp_server *server,
 	server_message_init(server, client_msg, reply);
 
 	_dhcp_message_builder_init(&builder, reply, len, DHCP_MESSAGE_TYPE_NAK);
-
+	copy_client_id(&builder, client_id);
 	_dhcp_message_builder_finalize(&builder, &len);
 
-	if (server->transport->l2_send(server->transport, server->address,
-					DHCP_PORT_SERVER, reply->ciaddr,
-					DHCP_PORT_CLIENT, MAC_BCAST_ADDR,
-					reply, len) < 0)
-		SERVER_DEBUG("Failed to send NACK");
+	server_message_send(server, reply, len, DHCP_MESSAGE_TYPE_NAK);
 }
 
 static void send_ack(struct l_dhcp_server *server,
-			const struct dhcp_message *client_msg, uint32_t dest)
+			const struct dhcp_message *client_msg,
+			struct l_dhcp_lease *lease,
+			bool rapid_commit)
 {
 	struct dhcp_message_builder builder;
 	size_t len = sizeof(struct dhcp_message) + DHCP_MIN_OPTIONS_SIZE;
 	L_AUTO_FREE_VAR(struct dhcp_message *, reply);
 	uint32_t lease_time = L_CPU_TO_BE32(server->lease_seconds);
-	struct l_dhcp_lease *lease;
+	L_AUTO_FREE_VAR(uint8_t *, client_id) = l_steal_ptr(lease->client_id);
 
 	reply = (struct dhcp_message *) l_new(uint8_t, len);
 
@@ -536,37 +681,39 @@ static void send_ack(struct l_dhcp_server *server,
 
 	_dhcp_message_builder_init(&builder, reply, len, DHCP_MESSAGE_TYPE_ACK);
 
-	reply->yiaddr = dest;
+	reply->yiaddr = lease->address;
 
 	_dhcp_message_builder_append(&builder,
 					L_DHCP_OPTION_IP_ADDRESS_LEASE_TIME,
 					4, &lease_time);
 
 	add_server_options(server, &builder);
+	copy_client_id(&builder, client_id);
 
 	_dhcp_message_builder_append(&builder, L_DHCP_OPTION_SERVER_IDENTIFIER,
 					4, &server->address);
 
+	if (rapid_commit)
+		_dhcp_message_builder_append(&builder, DHCP_OPTION_RAPID_COMMIT,
+						0, "");
+
 	_dhcp_message_builder_finalize(&builder, &len);
 
-	SERVER_DEBUG("Sending ACK to %s", IP_STR(reply->yiaddr));
+	SERVER_DEBUG("Sending ACK to "NIPQUAD_FMT, NIPQUAD(reply->yiaddr));
 
-	if (server->transport->l2_send(server->transport, server->address,
-					DHCP_PORT_SERVER, reply->ciaddr,
-					DHCP_PORT_CLIENT,
-					reply->chaddr, reply, len) < 0) {
-		SERVER_DEBUG("Failed to send ACK");
+	if (!server_message_send(server, reply, len, DHCP_MESSAGE_TYPE_ACK))
 		return;
-	}
 
-	lease = add_lease(server, false, reply->chaddr, reply->yiaddr);
+	lease = add_lease(server, false, client_id, reply->chaddr,
+				reply->yiaddr);
 
 	if (server->event_handler)
 		server->event_handler(server, L_DHCP_SERVER_EVENT_NEW_LEASE,
 					server->user_data, lease);
 }
 
-static void listener_event(const void *data, size_t len, void *user_data)
+static void listener_event(const void *data, size_t len, void *user_data,
+				const uint8_t *saddr)
 {
 	struct l_dhcp_server *server = user_data;
 	const struct dhcp_message *message = data;
@@ -575,11 +722,17 @@ static void listener_event(const void *data, size_t len, void *user_data)
 	const void *v;
 	struct l_dhcp_lease *lease;
 	uint8_t type = 0;
-	uint32_t server_id_opt = 0;
+	bool server_id_opt = false;
+	bool server_id_match = true;
 	uint32_t requested_ip_opt = 0;
+	L_AUTO_FREE_VAR(uint8_t *, client_id_opt) = NULL;
+	bool rapid_commit_opt = false;
 
 	SERVER_DEBUG("");
 
+	if (saddr && memcmp(saddr, message->chaddr, ETH_ALEN))
+		return;
+
 	if (!_dhcp_message_iter_init(&iter, message, len))
 		return;
 
@@ -591,61 +744,171 @@ static void listener_event(const void *data, size_t len, void *user_data)
 
 			break;
 		case L_DHCP_OPTION_SERVER_IDENTIFIER:
-			if (l == 4)
-				server_id_opt = l_get_u32(v);
-
-			if (server->address != server_id_opt)
-				return;
+			if (l != 4)
+				break;
 
+			server_id_opt = true;
+			server_id_match = (l_get_u32(v) == server->address);
 			break;
 		case L_DHCP_OPTION_REQUESTED_IP_ADDRESS:
 			if (l == 4)
 				requested_ip_opt = l_get_u32(v);
 
 			break;
+		case DHCP_OPTION_CLIENT_IDENTIFIER:
+			if (l < 1 || l > 253 || client_id_opt)
+				break;
+
+			client_id_opt = l_malloc(l + 1);
+			client_id_opt[0] = l;
+			memcpy(client_id_opt + 1, v, l);
+			break;
+		case DHCP_OPTION_RAPID_COMMIT:
+			rapid_commit_opt = true;
+			break;
 		}
 	}
 
 	if (type == 0)
 		return;
 
-	lease = find_lease_by_mac(server, message->chaddr);
+	if (requested_ip_opt)
+		lease = find_lease_by_id_and_ip(server->lease_list,
+						client_id_opt, message->chaddr,
+						requested_ip_opt);
+
+	if (!requested_ip_opt || !lease)
+		lease = find_lease_by_id(server->lease_list, client_id_opt,
+						message->chaddr);
+
 	if (!lease)
 		SERVER_DEBUG("No lease found for "MAC,
 					MAC_STR(message->chaddr));
 
 	switch (type) {
 	case DHCP_MESSAGE_TYPE_DISCOVER:
-		SERVER_DEBUG("Received DISCOVER, requested IP %s",
-					IP_STR(requested_ip_opt));
+		SERVER_DEBUG("Received DISCOVER, requested IP "NIPQUAD_FMT,
+					NIPQUAD(requested_ip_opt));
 
-		send_offer(server, message, lease, requested_ip_opt);
+		if (!server_id_match)
+			break;
 
+		if (rapid_commit_opt && server->rapid_commit) {
+			lease = l_dhcp_server_discover(server, requested_ip_opt,
+							client_id_opt,
+							message->chaddr);
+			if (!lease) {
+				send_nak(server, message, client_id_opt);
+				break;
+			}
+
+			send_ack(server, message, lease, rapid_commit_opt);
+			break;
+		}
+
+		send_offer(server, message, lease, requested_ip_opt,
+				client_id_opt);
 		break;
 	case DHCP_MESSAGE_TYPE_REQUEST:
-		SERVER_DEBUG("Received REQUEST, requested IP %s",
-				IP_STR(requested_ip_opt));
+		SERVER_DEBUG("Received REQUEST, requested IP "NIPQUAD_FMT,
+				NIPQUAD(requested_ip_opt));
 
-		if (requested_ip_opt == 0) {
-			requested_ip_opt = message->ciaddr;
-			if (requested_ip_opt == 0)
+		/*
+		 * RFC2131 Section 3.5: "Those servers not selected by the
+		 * DHCPREQUEST message use the message as notification that
+		 * the client has declined that server's offer."
+		 */
+		if (!server_id_match) {
+			if (server->authoritative) {
+				send_nak(server, message, client_id_opt);
 				break;
-		}
+			}
 
-		if (lease && requested_ip_opt == lease->address) {
-			send_ack(server, message, lease->address);
+			if (!lease || !lease->offering)
+				break;
+
+			remove_lease(server, lease);
 			break;
 		}
 
-		if (server_id_opt || !lease) {
-			send_nak(server, message);
+		/*
+		 * As an extension, check if we have an expired lease matching
+		 * the requested IP and the client ID/mac and if so, allow the
+		 * lease to be re-activated.
+		 */
+		if (!lease && requested_ip_opt)
+			lease = find_lease_by_id_and_ip(server->expired_list,
+							client_id_opt,
+							message->chaddr,
+							requested_ip_opt);
+
+		/*
+		 * RFC2131 Section 3.5: "If the selected server is unable to
+		 * satisfy the DHCPREQUEST message (...), the server SHOULD
+		 * respond with a DHCPNAK message."
+		 *
+		 * But:
+		 * 4.3.2: "If the DHCP server has no record of this client,
+		 * then it MUST remain silent (...)"
+		 */
+		if (!lease) {
+			if (server_id_opt || server->authoritative)
+				send_nak(server, message, client_id_opt);
+
 			break;
 		}
+
+		/*
+		 * 4.3.2: "If the DHCPREQUEST message contains a 'server
+		 * identifier' option, the message is in response to a
+		 * DHCPOFFER message.  Otherwise, the message is a request
+		 * to verify or extend an existing lease."
+		 */
+		if (server_id_opt && server_id_match) {
+			/*
+			 * Allow either no 'requested IP address' option or
+			 * a value identical with the one we offered because
+			 * the spec is unclear on whether it is to be
+			 * included:
+			 *
+			 * Section 4.3.2: "DHCPREQUEST generated during
+			 * SELECTING state: (...) 'requested IP address' MUST
+			 * be filled in with the yiaddr value from the chosen
+			 * DHCPOFFER."
+			 *
+			 * but section 3.5 suggests only in the INIT-REBOOT
+			 * state: "The 'requested IP address' option is to be
+			 * filled in only in a DHCPREQUEST message when the
+			 * client is verifying network parameters obtained
+			 * previously."
+			 */
+			if (!lease->offering ||
+					(requested_ip_opt &&
+					 requested_ip_opt != lease->address)) {
+				send_nak(server, message, client_id_opt);
+				break;
+			}
+		} else {
+			/*
+			 * 3.5: "If a server receives a DHCPREQUEST message
+			 * with an invalid 'requested IP address', the server
+			 * SHOULD respond to the client with a DHCPNAK message"
+			 */
+			if (lease->offering ||
+					(requested_ip_opt &&
+					 requested_ip_opt != lease->address)) {
+				send_nak(server, message, client_id_opt);
+				break;
+			}
+		}
+
+		send_ack(server, message, lease, false);
 		break;
 	case DHCP_MESSAGE_TYPE_DECLINE:
 		SERVER_DEBUG("Received DECLINE");
 
-		if (!server_id_opt || !requested_ip_opt || !lease)
+		if (!server_id_opt || !server_id_match || !requested_ip_opt ||
+				!lease)
 			break;
 
 		if (requested_ip_opt == lease->address)
@@ -655,7 +918,8 @@ static void listener_event(const void *data, size_t len, void *user_data)
 	case DHCP_MESSAGE_TYPE_RELEASE:
 		SERVER_DEBUG("Received RELEASE");
 
-		if (!server_id_opt || !lease)
+		if (!server_id_opt || !server_id_match || !lease ||
+				lease->offering)
 			break;
 
 		if (message->ciaddr == lease->address)
@@ -665,7 +929,10 @@ static void listener_event(const void *data, size_t len, void *user_data)
 	case DHCP_MESSAGE_TYPE_INFORM:
 		SERVER_DEBUG("Received INFORM");
 
-		send_inform(server, message);
+		if (!server_id_match)
+			break;
+
+		send_inform(server, message, client_id_opt);
 		break;
 	}
 }
@@ -705,13 +972,13 @@ struct dhcp_transport *_dhcp_server_get_transport(struct l_dhcp_server *server)
 LIB_EXPORT struct l_dhcp_server *l_dhcp_server_new(int ifindex)
 {
 	struct l_dhcp_server *server = l_new(struct l_dhcp_server, 1);
-	if (!server)
-		return NULL;
 
 	server->lease_list = l_queue_new();
 	server->expired_list = l_queue_new();
 
 	server->started = false;
+	server->authoritative = true;
+	server->rapid_commit = true;
 
 	server->lease_seconds = DEFAULT_DHCP_LEASE_SEC;
 	server->max_expired = MAX_EXPIRED_LEASES;
@@ -749,6 +1016,7 @@ LIB_EXPORT void l_dhcp_server_destroy(struct l_dhcp_server *server)
 
 LIB_EXPORT bool l_dhcp_server_start(struct l_dhcp_server *server)
 {
+	char buf[INET_ADDRSTRLEN];
 	struct in_addr ia;
 
 	if (unlikely(!server))
@@ -770,7 +1038,7 @@ LIB_EXPORT bool l_dhcp_server_start(struct l_dhcp_server *server)
 
 	/* Assign a default netmask if not already */
 	if (!server->netmask) {
-		if (inet_aton("255.255.255.0", &ia) < 0)
+		if (inet_pton(AF_INET,"255.255.255.0", &ia) != 1)
 			return false;
 
 		server->netmask = ia.s_addr;
@@ -778,13 +1046,45 @@ LIB_EXPORT bool l_dhcp_server_start(struct l_dhcp_server *server)
 
 	/*
 	 * Assign a default ip range if not already. This will default to
-	 * server->address + 1 ... 254
+	 * server->address + 1 ... subnet end address - 1
 	 */
 	if (!server->start_ip) {
-		server->start_ip = L_BE32_TO_CPU(server->address) + 1;
-		server->end_ip = (server->start_ip & 0xffffff00) | 0xfe;
+		server->start_ip = ntohl(server->address) + 1;
+		server->end_ip = (ntohl(server->address) |
+			(~ntohl(server->netmask))) - 1;
+	} else {
+		if ((server->start_ip ^ ntohl(server->address)) &
+				ntohl(server->netmask))
+			return false;
+
+		if ((server->end_ip ^ ntohl(server->address)) &
+				ntohl(server->netmask))
+			return false;
+
+		/*
+		 * Directly ensure the [start_ip, end_ip] range doesn't
+		 * include the subnet address or the broadcast address so that
+		 * we have fewer checks to make when selecting a free address
+		 * from that range.  Additionally this ensures end_ip is not
+		 * 0xffffffff so we can use the condition "<= server->end_ip"
+		 * safely on uint32_t values.
+		 * In find_free_or_expired_ip we skip over IPs ending in .0 or
+		 * .255 even for netmasks other than 24-bit just to avoid
+		 * generating addresses that could look suspicious even if
+		 * they're legal.  We don't exclude these addresses when
+		 * explicitly requested by the client, i.e. in
+		 * check_requested_ip.
+		 */
+		if ((server->start_ip & (~ntohl(server->netmask))) == 0)
+			server->start_ip++;
+
+		if ((server->end_ip | ntohl(server->netmask)) == 0xffffffff)
+			server->end_ip--;
 	}
 
+	if (server->start_ip >= server->end_ip)
+		return false;
+
 	if (!server->ifname) {
 		server->ifname = l_net_get_name(server->ifindex);
 
@@ -815,11 +1115,12 @@ LIB_EXPORT bool l_dhcp_server_start(struct l_dhcp_server *server)
 	l_acd_set_defend_policy(server->acd, L_ACD_DEFEND_POLICY_INFINITE);
 
 	ia.s_addr = server->address;
+	inet_ntop(AF_INET, &ia, buf, INET_ADDRSTRLEN);
 
 	/* In case of unit testing we don't want this to be a fatal error */
-	if (!l_acd_start(server->acd, inet_ntoa(ia))) {
+	if (!l_acd_start(server->acd, buf)) {
 		SERVER_DEBUG("Failed to start ACD on %s, continuing without",
-				IP_STR(server->address));
+				buf);
 
 		l_acd_destroy(server->acd);
 		server->acd = NULL;
@@ -866,12 +1167,12 @@ LIB_EXPORT bool l_dhcp_server_set_ip_range(struct l_dhcp_server *server,
 	if (unlikely(!server || !start_ip || !end_ip))
 		return false;
 
-	if (inet_aton((const char *)start_ip, &_host_addr) == 0)
+	if (inet_pton(AF_INET, start_ip, &_host_addr) != 1)
 		return false;
 
 	start = ntohl(_host_addr.s_addr);
 
-	if (inet_aton((const char *) end_ip, &_host_addr) == 0)
+	if (inet_pton(AF_INET, (const char *) end_ip, &_host_addr) != 1)
 		return false;
 
 	server->start_ip = start;
@@ -928,7 +1229,7 @@ LIB_EXPORT bool l_dhcp_server_set_ip_address(struct l_dhcp_server *server,
 	if (unlikely(!server))
 		return false;
 
-	if (inet_aton(ip, &ia) < 0)
+	if (inet_pton(AF_INET, ip, &ia) != 1)
 		return false;
 
 	server->address = ia.s_addr;
@@ -956,7 +1257,7 @@ LIB_EXPORT bool l_dhcp_server_set_netmask(struct l_dhcp_server *server,
 	if (unlikely(!server || !mask))
 		return false;
 
-	if (inet_aton(mask, &ia) < 0)
+	if (inet_pton(AF_INET, mask, &ia) != 1)
 		return false;
 
 	server->netmask = ia.s_addr;
@@ -972,7 +1273,7 @@ LIB_EXPORT bool l_dhcp_server_set_gateway(struct l_dhcp_server *server,
 	if (unlikely(!server || !ip))
 		return false;
 
-	if (inet_aton(ip, &ia) < 0)
+	if (inet_pton(AF_INET, ip, &ia) != 1)
 		return false;
 
 	server->gateway = ia.s_addr;
@@ -993,7 +1294,7 @@ LIB_EXPORT bool l_dhcp_server_set_dns(struct l_dhcp_server *server, char **dns)
 	for (i = 0; dns[i]; i++) {
 		struct in_addr ia;
 
-		if (inet_aton(dns[i], &ia) < 0)
+		if (inet_pton(AF_INET, dns[i], &ia) != 1)
 			goto failed;
 
 		dns_list[i] = ia.s_addr;
@@ -1010,3 +1311,160 @@ failed:
 	l_free(dns_list);
 	return false;
 }
+
+LIB_EXPORT void l_dhcp_server_set_authoritative(struct l_dhcp_server *server,
+						bool authoritative)
+{
+	if (unlikely(!server))
+		return;
+
+	server->authoritative = authoritative;
+}
+
+LIB_EXPORT void l_dhcp_server_set_enable_rapid_commit(
+						struct l_dhcp_server *server,
+						bool enable)
+{
+	if (unlikely(!server))
+		return;
+
+	server->rapid_commit = enable;
+}
+
+LIB_EXPORT struct l_dhcp_lease *l_dhcp_server_discover(
+						struct l_dhcp_server *server,
+						uint32_t requested_ip_opt,
+						const uint8_t *client_id,
+						const uint8_t *mac)
+{
+	struct l_dhcp_lease *lease;
+
+	SERVER_DEBUG("Requested IP " NIPQUAD_FMT " for " MAC,
+			NIPQUAD(requested_ip_opt), MAC_STR(mac));
+
+	if ((lease = find_lease_by_id(server->lease_list, client_id, mac)))
+		requested_ip_opt = lease->address;
+	else if (!check_requested_ip(server, requested_ip_opt)) {
+		requested_ip_opt = find_free_or_expired_ip(server, mac);
+
+		if (unlikely(!requested_ip_opt)) {
+			SERVER_DEBUG("Could not find any free addresses");
+			return NULL;
+		}
+	}
+
+	lease = add_lease(server, true, client_id, mac, requested_ip_opt);
+	if (unlikely(!lease)) {
+		SERVER_DEBUG("add_lease() failed");
+		return NULL;
+	}
+
+	SERVER_DEBUG("Offering " NIPQUAD_FMT " to " MAC,
+			NIPQUAD(requested_ip_opt), MAC_STR(mac));
+	return lease;
+}
+
+LIB_EXPORT bool l_dhcp_server_request(struct l_dhcp_server *server,
+					struct l_dhcp_lease *lease)
+{
+	uint8_t mac[ETH_ALEN];
+
+	if (unlikely(!lease))
+		return false;
+
+	SERVER_DEBUG("Requested IP " NIPQUAD_FMT " for " MAC,
+			NIPQUAD(lease->address), MAC_STR(lease->mac));
+
+	memcpy(mac, lease->mac, ETH_ALEN);
+	lease = add_lease(server, false, NULL, mac, lease->address);
+
+	if (server->event_handler)
+		server->event_handler(server, L_DHCP_SERVER_EVENT_NEW_LEASE,
+					server->user_data, lease);
+
+	return true;
+}
+
+LIB_EXPORT bool l_dhcp_server_decline(struct l_dhcp_server *server,
+					struct l_dhcp_lease *lease)
+{
+	if (unlikely(!lease || !lease->offering))
+		return false;
+
+	SERVER_DEBUG("Declined IP " NIPQUAD_FMT " for " MAC,
+			NIPQUAD(lease->address), MAC_STR(lease->mac));
+
+	return remove_lease(server, lease);
+}
+
+LIB_EXPORT bool l_dhcp_server_release(struct l_dhcp_server *server,
+					struct l_dhcp_lease *lease)
+{
+	if (unlikely(!lease || lease->offering))
+		return false;
+
+	SERVER_DEBUG("Released IP " NIPQUAD_FMT " for " MAC,
+			NIPQUAD(lease->address), MAC_STR(lease->mac));
+
+	lease_release(server, lease);
+	return true;
+}
+
+/* Drop an offered, active or expired lease without moving it to expired_list */
+LIB_EXPORT bool l_dhcp_server_lease_remove(struct l_dhcp_server *server,
+						struct l_dhcp_lease *lease)
+{
+	if (unlikely(!lease))
+		return false;
+
+	if (unlikely(!l_queue_remove(server->lease_list, lease) &&
+			!l_queue_remove(server->expired_list, lease)))
+		return false;
+
+	_dhcp_lease_free(lease);
+	set_next_expire_timer(server, NULL);
+	return true;
+}
+
+struct dhcp_expire_by_mac_data {
+	struct l_dhcp_server *server;
+	const uint8_t *mac;
+	unsigned int expired_cnt;
+};
+
+static bool dhcp_expire_by_mac(void *data, void *user_data)
+{
+	struct l_dhcp_lease *lease = data;
+	struct dhcp_expire_by_mac_data *expire_data = user_data;
+	struct l_dhcp_server *server = expire_data->server;
+
+	if (!match_lease_mac(lease, expire_data->mac))
+		return false;
+
+	if (server->event_handler)
+		server->event_handler(server, L_DHCP_SERVER_EVENT_LEASE_EXPIRED,
+					server->user_data, lease);
+
+	if (!lease->offering) {
+		if (l_queue_length(server->expired_list) > server->max_expired)
+			_dhcp_lease_free(l_queue_pop_head(server->expired_list));
+
+		l_queue_push_tail(server->expired_list, lease);
+	} else
+		_dhcp_lease_free(lease);
+
+	expire_data->expired_cnt++;
+	return true;
+}
+
+LIB_EXPORT void l_dhcp_server_expire_by_mac(struct l_dhcp_server *server,
+						const uint8_t *mac)
+{
+	struct dhcp_expire_by_mac_data expire_data = { server, mac, 0 };
+
+	l_queue_foreach_remove(server->lease_list, dhcp_expire_by_mac,
+				&expire_data);
+
+	if (expire_data.expired_cnt)
+		set_next_expire_timer(server, NULL);
+}
